import os
import pathlib
import shutil
import sys
import uuid
import glob
import re
import threading
from datetime import datetime

__author__ = "Mainul Hossain"
__date__ = "$Dec 10, 2016 2:23:14 PM$"

class FolderItem():
    def __init__(self, path):
        self.path = path
    
    def __str__(self):
        return self.path
    
    def __repr__(self):
        return self.__str__()

    def __add__(self, other):
        return FolderItem.union(self.path, other)
    
    def __sub__(self, other):
        return FolderItem.substract(self.path, other)
        
    @staticmethod
    def StrToFolderItem(item_s):
        return [FolderItem(f) for f in item_s] if isinstance(item_s, list) else FolderItem(item_s)
    
#     @staticmethod
#     def union(left, right):
#         v = []
#         if not isinstance(left, list):
#             left = [left]
#             
#         if not isinstance(right, list):
#             right = [right]
#         
#         rest = []
#         for l in left:
#             rest = [r for r in right if str(l) != str(r)]
#         left.extend(rest)
#         return left
    
    # set operations  
    @staticmethod
    def union(left, right):
        if not isinstance(left, list):
            left = [left]
             
        if not isinstance(right, list):
            right = [right]
         
        left = [str(f) for f in left]
        right = [str(f) for f in right]
        left_s = set(left)
        right_s = set(right)
         
        left = left_s.union(right_s)
        return [FolderItem(f) for f in left]
      
#     @staticmethod
#     def substract(left, right):
#         if not isinstance(left, list):
#             left = [left]
#             
#         if not isinstance(right, list):
#             right = [right]
#             
#         for r in right:
#             left = [l for l in left if str(l) != str(r)]
#             
#         return left
    
    
    @staticmethod
    def substract(left, right):
        if not isinstance(left, list):
            left = [left]
             
        if not isinstance(right, list):
            right = [right]
         
        left = [str(f) for f in left]
        right = [str(f) for f in right]
        left_s = set(left)
        right_s = set(right)
         
        left = left_s - right_s
        return [FolderItem(f) for f in left]                    
    
class FilterManager:
    @staticmethod
    def Filter(fs, path, filename, filters):
        for fil in filters:
            if not fil['selected']:
                continue
            name = fil['name'].lower()
            if name == "name":
                if not FilterManager.CheckRegExValue(fil['value'], filename):
                    return None
            elif name == "created" or name == "modified" or name == "accessed" or name == "size":
                stat = fs.stat(fs.join(path, filename))
                if name == "size":
                    if not FilterManager.SizeInRange(stat.st_size, fil['value']):
                        return None
                else:
                    t = None
                    if name == "created":
                        t = stat.st_ctime
                    elif name == "modified":
                        t = stat.st_mtime
                    else:
                        t = stat.st_atime
                    if not FilterManager.DateInRange(t, fil['value']):
                        return None
            elif name == "key:value":
                pass
            
        return filename

    @staticmethod
    def CheckRegExValue(regexvalue, value):
        return re.search(regexvalue, value, flags=re.RegexFlag.IGNORECASE)
        
    @staticmethod
    def DateInRange(dateV, dateR):
        begin_end = FilterManager.SplitDateRange(dateR)
        if not begin_end:
            return True
        if dateV < datetime.timestamp(begin_end[0]):
            return False
        if len(begin_end) > 1 and  dateV > datetime.timestamp(begin_end[1]):
            return False
        return True
    
    @staticmethod
    def SizeInRange(sizeV, sizeR):
        sizeR = sizeR.replace(" KB", "")
        begin_end = sizeR.split(",")
        if not begin_end:
            return True
        if sizeV < int(begin_end[0]) * 1024:
            return False
        if len(begin_end) > 1 and  sizeV > int(begin_end[1]) * 1024:
            #print_len = len(begin_end)
            #print_endSize = int(begin_end[1]) * 1024
            return False
        return True
    
    @staticmethod
    def SplitDateRange(dateR):
        begin_end = dateR.split('-')
        return [datetime.strptime(begin_end[0].strip(), '%m/%d/%Y'), datetime.strptime(begin_end[1].strip(), '%m/%d/%Y')] 

    @staticmethod
    def Sort(fs, filenames, filters):
        for fil in filters:
            if fil['selected'] and fil['sort']:
                name = fil['name'].lower()
                if name == "name":
                    filenames = sorted(filenames, reverse = (int(fil['sort'])==2))
        return filenames
    
    @staticmethod
    def listdirR(fs, path, filters, flatfiles = False):
        
        ancestors = FilterManager.listdirRInternal(fs, path, filters, flatfiles)
        if not ancestors:
            return None
        
        if flatfiles:
            return ancestors
        
        ancestorPath = path
        while fs.dirname(ancestorPath) and fs.strip_root(ancestorPath) != os.sep:
            ancestorPath = fs.dirname(ancestorPath) 
            ancestor = fs.make_json_item(ancestorPath)
            if fs.strip_root(ancestorPath) == os.sep:
                ancestor['text'] = fs.prefix
            ancestor['children'].append(ancestors)
            ancestors = ancestor
                
        return ancestors
            
    @staticmethod
    def listdirRInternal(fs, path, filters, flatfiles = False):
        if fs.isfile(path):
            filtered_file = filter(path, filters)
            if not filtered_file:
                return None
            
            return fs.make_json_item(path) if not flatfiles else path

        fsitems = []
        for fd in fs.get_folders(path):
            fditem = FilterManager.listdirRInternal(fs, fs.join(path, fd), filters, flatfiles)
            if fditem:
                fsitems.append(fditem)
        
        filter_active = any(f['selected'] for f in filters)
        fitems = []
        for f in fs.get_files(path):
            filename = fs.basename(f)
            if not filter_active or FilterManager.Filter(fs, path, filename, filters):
                fitems.append(filename)
                
        if fitems:
            fitems = FilterManager.Sort(fs, fitems, filters)
            for fitem in fitems:
                if flatfiles:
                    fsitems.append(fs.join(path, fitem))
                else:
                    fsitems.append(fs.make_json_item(fs.join(path, fitem)))
        
        if fsitems and not flatfiles:
            data_json = fs.make_json_item(path)
            data_json['children'] = fsitems
            data_json['loaded'] = True
            fsitems = data_json
            
        return fsitems
        
class BaseFileSystem(object):
    def __init__(self, root, public, temp, prefix):
        self.__lock = threading.Lock()
        self._localdir = root
        self.prefix = prefix
        self.public = public
        self.url = '/'
        self.temp = temp
    
    @property
    def localdir(self):
        return self._localdir
    @localdir.setter
    def localdir(self, dir):
        self.__lock.acquire()
        try:
            self._localdir = dir
        finally:
            self.__lock.release()
        
    def is_parent_of(self, parent, child):
        parent = self.normalize_path(parent)
        child = self.normalize_path(child)
        return child.startswith(parent)

    def normalize_fullpath(self, path):
        return self.normalize_path(path)
    
    def mkdir(self, path):
        raise ValueError("Not implemented error.")
    
    def remove(self, path):
        raise ValueError("Not implemented error.")
    
    def rename(self, oldpath, newpath):
        raise ValueError("Not implemented error.")
    
    def get_files(self, path, pattern = None, recursive = False):
        raise ValueError("Not implemented error.")
    
    def get_folders(self, path, pattern = None, recursive = False):
        raise ValueError("Not implemented error.")
    
    def listdir(self, path):
        raise ValueError("Not implemented error.")
    
    def listdirR(self, path, filters):
        raise ValueError("Not implemented error.")
    
    def copyfile(self, src, dst):
        raise ValueError("Not implemented error.")
                
    def read(self, path):
        raise ValueError("Not implemented error.")
    
    def write(self, path, content):
        raise ValueError("Not implemented error.")
        
    def unique_filename(self, path, prefix, ext):
        raise ValueError("Not implemented error.")
    
    def make_unique_dir(self, path):
        raise ValueError("Not implemented error.")
            
    def exists(self, path):
        raise ValueError("Not implemented error.")
        
    def isdir(self, path):
        raise ValueError("Not implemented error.")
    
    def isfile(self, path):
        raise ValueError("Not implemented error.")
    
    def join(self, path1, path2):
        raise ValueError("Not implemented error.")
    
    def basename(self, path):
        raise ValueError("Not implemented error.")
    
    #check again
    def dirname(self, path):
        raise ValueError("Not implemented error.")

    def make_json(self, path):
        raise ValueError("Not implemented error.")
    
    def make_json_r(self, path):
        raise ValueError("Not implemented error.")

    def save_upload(self, file, path, offset=None):
        raise ValueError("Not implemented error.")
    
    def download(self, path):
        raise ValueError("Not implemented error.")
    
    def stat(self, path):
        raise ValueError("Not implemented error.")

    def unique_fs_name(self, path, prefix, ext):
        make_fn = lambda i: os.path.join(path, '{0}({1}){2}'.format(prefix, i, ext))

        for i in range(1, sys.maxsize):
            uni_fn = make_fn(i)
            if not self.exists(uni_fn):
                return uni_fn

class PosixFileSystem(BaseFileSystem):
    
    def __init__(self, root, public = None, temp = None, prefix = None):
        super().__init__(root, public, temp, prefix)
    
    def typename(self):
        return "posix"
        
    def normalize_path(self, path):
        path = os.path.normpath(path)
        if self.prefix:
            path = self.strip_prefix(path)
        if not self._localdir or path.startswith(self._localdir) or (self._localdir.endswith(os.sep) and path == self._localdir[:-1]):
            return path
        while path and path[0] == os.sep:
            path = path[1:]    
        return os.path.join(self._localdir, path)
    
    def makedirs(self, path):
        path = self.normalize_path(path)
        if not os.path.exists(path):
            os.makedirs(path)
        return self.make_prefix(path)
    
    def make_prefix(self, path):
        if not self.prefix:
            return path     
        if path.startswith(self.prefix):
            return path
        if path.startswith(self._localdir):
            path = path[len(self._localdir):]
        if not path.startswith(os.sep):
            path = os.sep + path
        return self.prefix + path
    
    def make_url(self, path):
        path = self.strip_prefix(path) 
        if self._localdir and path.startswith(self._localdir):
            path = path[len(self._localdir):]
        return path if path.startswith(os.sep) else os.sep + path
    
    def strip_prefix(self, path):
        return path[len(self.prefix):] if self.prefix and path.startswith(self.prefix) else path
        
    def strip_root(self, path):
        path = self.strip_prefix(path)
        if path.startswith(self._localdir):
            path = path[len(self._localdir):]
        return path if path.startswith(os.sep) else os.sep + path
            
    def mkdir(self, path):
        path = self.normalize_path(path)
        if not os.path.exists(path):
            os.mkdir(path) 
        return self.make_prefix(path)
    
    def remove(self, path):
        path = self.normalize_path(path)
        dirpath = os.path.dirname(path)
        if os.path.isdir(path):
            shutil.rmtree(path)
        elif os.path.isfile(path):
            os.remove(path)
        return self.make_prefix(dirpath)
               
    def rename(self, oldpath, newpath):
        oldpath = self.normalize_path(oldpath)
        newpath = self.join(self.dirname(oldpath), newpath)
        newpath = self.normalize_path(newpath)
        
        if oldpath == newpath:
            return oldpath
        
        os.rename(oldpath, newpath)
        return self.strip_root(newpath)
    
    def __filesOrFolders(self, path, pattern = None, recursive = False, folder = False):
        if not pattern:
            pattern = '*'
        if recursive:
            pattern = '**' + os.sep + pattern
        path = self.normalize_path(path)
        path = path + os.sep + pattern
        if folder and not path.endswith(os.sep):
            path = path + os.sep
        if not folder and path.endswith(os.sep):
            path = path[-1]

        return [self.strip_root(f) for f in glob.glob(path, recursive=recursive) if folder or self.isfile(f)]
    
    def get_files(self, path, pattern = None, recursive = False):
        return self.__filesOrFolders(path, pattern, recursive, False)
    
    def get_folders(self, path, pattern = None, recursive = False):
        return self.__filesOrFolders(path, pattern, recursive, True)

    def listdir(self, path):
        path = self.normalize_path(path)
        return os.listdir(path)
    
    def listdirR(self, path, filters):
        filters = filters if not filters or type(filters) is list else [filters]
        return FilterManager.listdirR(self, path, filters)
       
    def copyfile(self, src, dst):
        return shutil.copy2(self.normalize_path(src), self.normalize_path(dst))
    
    def stat(self, path):
        return os.stat(self.normalize_path(path))
    
    def read(self, path):
        path = self.normalize_path(path)
        with open(path, 'rb') as reader:
            return reader.read()
        
    def write(self, path, content):
        path = self.normalize_path(path)
        with open(path, 'wb') as writer:
            return writer.write(content)
        
    def unique_filename(self, path, prefix, ext):
        stem = pathlib.Path(prefix).stem
        if not ext:
            ext = pathlib.Path(prefix).suffix
        make_fn = lambda i: self.join(path, '{0}_{1}.{2}'.format(stem, i, ext) if ext else '{0}_{1}'.format(stem, i))

        for i in range(1, sys.maxsize):
            uni_fn = make_fn(i)
            if not self.exists(uni_fn):
                return uni_fn
    
    def make_unique_dir(self, path):
        unique_dir = self.join(path, str(uuid.uuid4()))
        unique_dir = self.normalize_path(unique_dir)
        os.makedirs(unique_dir)
        return unique_dir
            
    def exists(self, path):
        return os.path.exists(self.normalize_path(path))
        
    def isdir(self, path):
        return os.path.isdir(self.normalize_path(path))
    
    def isfile(self, path):
        return os.path.isfile(self.normalize_path(path))
    
    def join(self, path1, path2):
        path1 = self.normalize_path(path1)
        return self.make_url(os.path.join(path1, path2))
    
    def get_json_name(self, path):
        if (self._localdir and self._localdir == path) or path == os.sep:
            return self.prefix if self.prefix else self._localdir 
        
        return self.basename(path)
        
    def make_json_item(self, path):
        data_json =  { 'path': self.make_url(path), 'text': self.get_json_name(path) }
        if self.isdir(path):
            data_json['children'] = []
            data_json['type'] = 'folder'
        else:
            data_json['type'] = 'file'
        return data_json
        
    def make_json(self, path):
        data_json = self.make_json_item(path)
        
        if 'children' in data_json: # folder
            data_json['children'] = [self.make_json_item(self.join(path, fn)) for fn in self.listdir(path)]
            data_json['loaded'] = True
        return data_json
    
    def make_json_r(self, path):
        data_json = self.make_json_item(path)  
        if 'children' in data_json: # folder 
            data_json['children'] = [self.make_json_r(self.join(path, fn)) for fn in self.listdir(path)]
            data_json['loaded'] = True
        return data_json

    def save_upload(self, file, path, offset=None):
        path = self.normalize_path(path)
        if self.isfile(path):
            path = os.path.dirname(path)
        elif not os.path.exists(path):
            os.makedirs(path)
        path = os.path.join(path, file.filename)   
        if offset is None:
            file.save(path)
        else:
            try:
                with open(path, 'ab') as f:
                    f.seek(offset)
                    f.write(file.stream.read())
            except OSError:
            # log.exception will include the traceback so we can see what's wrong
                return
        return self.strip_root(path), path
        
    def download(self, path):
        path = self.normalize_path(path)
        return path if os.path.isfile(path) else None
    
    def basename(self, path):
        path = self.normalize_path(path)
        return os.path.basename(path)
    
    #check again
    def dirname(self, path):
        path = self.strip_root(path)
        return os.path.dirname(path) if path.startswith('/') else path
    
if __name__ == "__main__":
    fs = PosixFileSystem('/root')
    files = fs.get_files('/wfdsl', pattern='*.py')
    for f in files:
        print(f)


parallel:
    bio_data = 'bio_data'
with:
    img_data = 'img_data'
with:
    text_data = 'text_data'

print(bio_data)
print(img_data)
print(text_data)

task process_data(data_type):
    if data_type == 'bio':
        print("Processing bio_data")
    cinky = "cinky"
    print(cinky)

# Function definition
# task process_data(data_type):
#     # global bio_data, img_data, text_data
#     if data_type == 'bio':
#         print(f"Processing {bio_data}")
#     elif data_type == 'img':
#         print(f"Processing {img_data}")
#     else:
#         print(f"Processing {text_data}")

# For loop
data_types = ['bio', 'img', 'text']
# for data_type in data_types:
#     process_data(data_type)
#     # process_data(data_type)
# print(cinky)

parfor data_type in data_types:
    print('cinky')




import logging
import threading
import _thread
from .context import Context


class CodeGenerator(object):
    '''
    The code generator for PhenoWL DSL
    '''
    def __init__(self):
        self.context = Context()
        self.code = ''
        self.imports = set()
        self.indent = 0

    def get_params(self, expr):
        v = []
        for e in expr:
            v.append(self.eval(e))
        return v
            
    def indent_stmt(self, str):
        return " " * self.indent + str
    
    def dofunc(self, expr):
        '''
        Execute func expression.
        :param expr:
        '''
        function = expr[0] if len(expr) < 3 else expr[1]
        package = expr[0][:-1] if len(expr) > 2 else None
        
        params = expr[1] if len(expr) < 3 else expr[2]
        v = self.get_params(params)
        
        # call task if exists
        if package is None and function in self.context.library.tasks:
            return self.context.library.code_run_task(function, v, self.dotaskstmt)

        if not self.context.library.check_function(function, package):
            raise Exception(r"'{0}' doesn't exist.".format(function))
            
        return self.context.library.code_func(self.context, package, function, v)
    
    def dorelexpr(self, expr):
        '''
        Executes relative expression.
        :param expr:
        '''
        left = self.eval(expr[0])
        right = self.eval(expr[2])
        operator = expr[1]
        if operator == '<':
            return "{0} < {1}".format(str(left), str(right))
        elif operator == '>':
            return "{0} > {1}".format(str(left), str(right))
        elif operator == '<=':
            return "{0} <= {1}".format(str(left), str(right))
        elif operator == '>=':
            return "{0} >= {1}".format(str(left), str(right))
        else:
            return "{0} == {1}".format(str(left), str(right))
    
    def doand(self, expr):
        '''
        Executes "and" expression.
        :param expr:
        '''
        if expr is None:#if expr is empty:
            return True
        right = self.eval(expr[-1])
        if len(expr) == 1:
            return right
        left = expr[:-2]
        if len(left) > 1:
            left = ['ANDEXPR'] + left
        left = self.eval(left)
        return "{0} and {1}".format(str(left), str(right))
    
    def dopar(self, expr):
        code = 'taskManager = TaskManager()\n'
#         for stmt in expr:
#             code += 'taskManager.submit_func(lambda: ' + self.eval(stmt) + ')\n'
        return code
    
    def dopar_stmt(self, expr):
        '''
        Execute a parallel expression.
        :param expr:
        '''
        self.run_multstmt(lambda: self.eval(expr))
    
    def run_multstmt(self, f):
        return f()

    def dolog(self, expr):
        '''
        Executes a logical expression.
        :param expr:
        '''
        right = self.eval(expr[-1])
        if len(expr) == 1:
            return right
        left = expr[:-2]
        if len(left) > 1:
            left = ['LOGEXPR'] + left
        left = self.eval(left)
        return "{0} or {1}".format(str(left), str(right))
    
    def domult(self, expr):
        '''
        Executes a multiplication/division operation
        :param expr:
        '''
        right = self.eval(expr[-1])
        if len(expr) == 1:
            return right
        left = expr[:-2]
        if len(left) > 1:
            left = ['MULTEXPR'] + left
        left = self.eval(left)
        return "{0} / {1}".format(str(left), str(right)) if expr[-2] == '/' else "{0} * {1}".format(str(left), str(right))

    def doarithmetic(self, expr):
        '''
        Executes arithmetic operation.
        :param expr:
        '''
        right = self.eval(expr[-1])
        if len(expr) == 1:
            return right
        left = expr[:-2]
        if len(left) > 1:
            left = ['NUMEXPR'] + left
        left = self.eval(left)
        return "{0} + {1}".format(str(left), str(right)) if expr[-2] == '+' else "{0} - {1}".format(str(left), str(right))
    
    def doif(self, expr):
        '''
        Executes if statement.
        :param expr:
        '''
        code = "if " + self.eval(expr[0]) + ":\n"
        code += self.run_multstmt(lambda: self.eval(expr[1]))
        if len(expr) > 3:
            code += "else:\n"
            code += self.run_multstmt(lambda: self.eval(expr[3]))
        return code
    
    def dolock(self, expr):
        if not self.context.symtab.var_exists(expr[0]) or not isinstance(self.context.symtab.get_var(expr[0]), _thread.RLock):
            self.context.symtab.add_var(expr[0], threading.RLock())    
        with self.context.symtab.get_var(expr[0]):
            self.eval(expr[1])
        pass
        
    def doassign(self, expr):
        '''
        Evaluates an assignment expression.
        :param expr:
        '''
        return "{0} = {1}".format(expr[0], self.eval(expr[1]))
        
    def dofor(self, expr):
        '''
        Execute a for expression.
        :param expr:
        '''
        code = "for {0} in {1}:\n".format(self.eval(expr[0]), self.eval(expr[1]))
        code += self.run_multstmt(lambda: self.eval(expr[2]))
        return code
    
    def eval_value(self, str_value):
        '''
        Evaluate a single expression for value.
        :param str_value:
        '''
        return str_value
    
    def dolist(self, expr):
        '''
        Executes a list operation.
        :param expr:
        '''
        v = []
        for e in expr:
            v.append(self.eval(e))
        return v
    
    def dolistidx(self, expr):
        val = self.context.get_var(expr[0])
        return val[self.eval(expr[1])]
    
    def dostmt(self, expr):
        if len(expr) > 1:
            logging.debug("Processing line: {0}".format(expr[0]))
            self.line = int(expr[0])
            return self.indent_stmt(self.eval(expr[1:])) + '\n'

    def dotaskdefstmt(self, expr):
        if not expr[0]:
            v = self.get_params(expr[1])
            return self.dotaskstmt(expr, v)
        else:
            self.context.library.add_task(expr[0], expr)
            return ''
    
    def dotaskstmt(self, expr, args):
        server = args[0] if len(args) > 0 else None
        user = args[1] if len(args) > 1 else None
        password = args[2] if len(args) > 2 else None
        
        if not server:
            server = self.eval(expr[1][0]) if len(expr[1]) > 0 else None
        if not user:
            user = self.eval(expr[1][1]) if len(expr[1]) > 1 else None
        if not password:
            password = self.eval(expr[1][2]) if len(expr[1]) > 2 else None
        
        try:
            self.context.append_dci(server, user, password)
            return 'if True:\n' + self.eval(expr[2])
        finally:
            self.context.pop_dci()
                    
    def eval(self, expr):        
        '''
        Evaluate an expression
        :param expr: The expression in AST tree form.
        '''
        if not isinstance(expr, list):
            return self.eval_value(expr)
        if not expr:
            return
        if len(expr) == 1:
            return self.eval(expr[0])
        if expr[0] == "FOR":
            return self.dofor(expr[1])
        elif expr[0] == "ASSIGN":
            return self.doassign(expr[1:])
        elif expr[0] == "CONST":
            return self.eval_value(expr[1])
        elif expr[0] == "NUMEXPR":
            return self.doarithmetic(expr[1:])
        elif expr[0] == "MULTEXPR":
            return self.domult(expr[1:])
        elif expr[0] == "CONCAT":
            return self.doarithmetic(expr[1:])
        elif expr[0] == "LOGEXPR":
            return self.dolog(expr[1:])
        elif expr[0] == "ANDEXPR":
            return self.doand(expr[1:])
        elif expr[0] == "RELEXPR":
            return self.dorelexpr(expr[1:])
        elif expr[0] == "IF":
            return self.doif(expr[1])
        elif expr[0] == "LIST":
            return self.dolist(expr[1])
        elif expr[0] == "FUNCCALL":
            code, imports = self.dofunc(expr[1])
            self.imports.update(imports)
            return code
        elif expr[0] == "LISTIDX":
            return self.dolistidx(expr[1])
        elif expr[0] == "PAR":
            return self.dopar(expr[1])
        elif expr[0] == "LOCK":
            return self.dolock(expr[1:])
        elif expr[0] == "STMT":
            return self.dostmt(expr[1:])
        elif expr[0] == "TASK":
            return self.dotaskdefstmt(expr[1:])
        elif expr[0] == "MULTISTMT":
            try:
                self.indent = int(expr[1].pop()) - 1
                return self.eval(expr[2:])
            finally:
                self.indent = int(expr[1].pop()) - 1
        else:
            code = ''
            for subexpr in expr:
                code += self.eval(subexpr)
            return code

    # Run it
    def run(self, prog):
        '''
        Run a new program.
        :param prog: Pyparsing ParseResults
        '''
        try:
            self.context.reload()
            stmt = prog.asList()
            code = self.eval(stmt)
            imports = ''
            for i in self.imports:
                imports = i + '\n';
            self.context.out = imports + '\n' + code 
        except Exception as err:
            self.context.err.append("Error at line {0}: {1}".format(self.line, err))



import threading
from dsl.symtab import SymbolTable

class Context:
    '''
    The context for parsing and interpretation.
    '''
    def __init__(self, library, SymbolTableCls=SymbolTable):
        '''
        Initializes this object.
        '''
        self.library = library
        self.runnable = None
        self.user_id = None
        self.SymbolTableCls = SymbolTableCls # custom symbol table from the client
        self.lock = threading.Lock()
        self.reload()
            
    def reload(self):
        """

        This method resets the following attributes:
        - symtab_stack: Dictionary to store symbol tables.
        - out: List to store output messages.
        - err: List to store error messages.
        - dci: List to store data context information.
        - globals: Dictionary to store global variables.
        - ident: Thread identifier obtained from threading.get_ident().
        - symtabs: List containing a new instance of SymbolTableCls, thread-specific symbol tables.
        """
        '''
        Reinitializes this object for new processing.
        '''
        self.out = []
        self.err = []
        self.dci = []
        self.globals = {}
        self.symtab_stack = {threading.get_ident(): [self.SymbolTableCls()]} # {thread_id: [SymbolTableCls(), SymbolTableCls()]}
        
    def get_var(self, name):
        '''
        Gets a variable from symbol stack and symbol table
        :param name:
        '''
        if not self.var_exists(name):
            raise ValueError(f"Variable {name} does not exist in the symbol table.")

        current_thread = threading.get_ident()
        with self.lock:
            if current_thread in self.symtab_stack:
                for s in reversed(self.symtab_stack[current_thread]):
                    if s.var_exists(name):
                        return s.get_var(name)        
    
    def add_var(self, name, value):
        """
        Adds a variable to the symbol table.
        If the variable already exists, it updates the variable with the new value.
        If the current thread matches the main thread, it adds the variable to the last symbol table in the stack.
        If the current thread does not have a symbol table stack, it creates one and adds the variable to the new symbol table.
        Args:
            name (str): The name of the variable to add.
            value (Any): The value of the variable to add.
        Returns:
            Any: The result of adding or updating the variable in the symbol table.
        """
        if self.var_exists(name):
            raise ValueError(f"Variable {name} already exists in the symbol table.")

        # if the current thread is the main thread
        current_thread = threading.get_ident()

        with self.lock:
            if not current_thread in self.symtab_stack:
                self.symtab_stack[current_thread] = [self.SymbolTableCls()]
            return self.symtab_stack[current_thread][-1].add_var(name, value)
            
    def update_var(self, name, value):
        """
        Update the value of a variable in the symbol table.

        This method searches for the variable in the symbol tables associated with the current thread.
        If the variable is found, its value is updated. The search is performed in reverse order of the symbol tables.

        Args:
            name (str): The name of the variable to update.
            value: The new value to assign to the variable.

        Returns:
            The result of the update operation from the symbol table, if the variable is found.
            Otherwise, returns None.
        """
        if not self.var_exists(name):
            raise ValueError(f"Variable {name} does not exist in the symbol table.")

        current_thread = threading.get_ident()

        with self.lock:
            if current_thread in self.symtab_stack:
                for s in reversed(self.symtab_stack[current_thread]):
                    if s.var_exists(name):
                        return s.update_var(name, value)
    
    def add_or_update_var(self, name, value):
        if self.var_exists(name):
            return self.update_var(name, value)
        else:
            return self.add_var(name, value)
                                
    def var_exists(self, name):
        '''
        Checks if a variable exists in any of the symbol tables.
        :param name: variable name
        '''
        current_thread = threading.get_ident()
        if current_thread in self.symtab_stack:
            for s in reversed(self.symtab_stack[current_thread]):
                if s.var_exists(name):
                    return True
    
    def append_local_symtab(self):
        '''
        Appends a new symbol table to the symbol table stack.
        '''
        with self.lock:
            current_thread = threading.get_ident()
            if current_thread in self.symtab_stack: # if current thread exists in the symbol table stack, append a new symbol table at the end
                self.symtab_stack[current_thread].append(self.SymbolTableCls())
            else: # otherwise, create a new symbol table stack
                self.symtab_stack[current_thread] = [self.SymbolTableCls()]
            return self.symtab_stack[current_thread][len(self.symtab_stack[current_thread]) - 1]
    
    def pop_local_symtab(self):
        '''
        Pop a symbol table from the symbol table stack.
        '''
        with self.lock:
            current_thread = threading.get_ident()
            if current_thread in self.symtab_stack:
                if self.symtab_stack[current_thread]:
                    self.symtab_stack[current_thread].pop()
                    if not self.symtab_stack[current_thread]: # no symbol table, remove the entry
                        del self.symtab_stack[current_thread]
        
    @property
    def library(self):
        return self.__library
    
    @library.setter
    def library(self, lib):
        self.__library = lib
        
    @property
    def runnable(self):
        return self.__runnable
    
    @runnable.setter
    def runnable(self, runnable):
        self.__runnable = runnable
                      
    def iequal(self, str1, str2):
        '''
        Compares two strings for case insensitive equality.
        :param str1:
        :param str2:
        '''
        if str1 == None:
            return str2 == None
        if str2 == None:
            return str1 == None
        return str1.lower() == str2.lower()
    
    def write(self, *args):
        '''
        Writes a line of strings in out context.
        '''
        self.out.append("{0}".format(', '.join(map(str, args))))
    
    def error(self, *args):
        '''
        Writes a line of strings in err context.
        '''
        self.err.append("{0}".format(', '.join(map(str, args))))

    def append_dci(self, server, user, password):
        self.dci.append([server, user, password])
    
    def pop_dci(self):
        if self.dci:
            return self.dci.pop()
    
    def get_activedci(self):
        if not self.dci:
            return [None, None, None]
        return self.dci[-1]